@@ -54,7 +54,6 @@ function SchemaType(path, options, instance) {
     }
   }
 
-
   if (options.select == null) {
     delete options.select;
   }
@@ -442,6 +441,38 @@ SchemaType.prototype.immutable = function(bool) {
   return this;
 };
 
+/**
+ * Defines a custom function for transforming this path when converting a document to JSON.
+ *
+ * Mongoose calls this function with one parameter: the current `value` of the path. Mongoose
+ * then uses the return value in the JSON output.
+ *
+ * ####Example:
+ *
+ *     const schema = new Schema({
+ *       date: { type: Date, transform: v => v.getFullYear() }
+ *     });
+ *     const Model = mongoose.model('Test', schema);
+ *
+ *     await Model.create({ date: new Date('2016-06-01') });
+ *     const doc = await Model.findOne();
+ *
+ *     doc.date instanceof Date; // true
+ *
+ *     doc.toJSON().date; // 2016 as a number
+ *     JSON.stringify(doc); // '{"_id":...,"date":2016}'
+ *
+ * @param {Function} fn
+ * @return {SchemaType} this
+ * @api public
+ */
+
+SchemaType.prototype.transform = function(fn) {
+  this.options.transform = fn;
+
+  return this;
+};
+
 /**
  * Adds a setter to this schematype.
  *
@@ -929,8 +960,10 @@ SchemaType.prototype.required = function(required, message) {
  *     const User = mongoose.model('User', userSchema);
  *
  *     const postSchema = new Schema({ user: mongoose.ObjectId });
- *     postSchema.path('user').ref('User'); // By model name
- *     postSchema.path('user').ref(User); // Can pass the model as well
+ *     postSchema.path('user').ref('User'); // Can set ref to a model name
+ *     postSchema.path('user').ref(User); // Or a model class
+ *     postSchema.path('user').ref(() => 'User'); // Or a function that returns the model name
+ *     postSchema.path('user').ref(() => User); // Or a function that returns the model class
  *
  *     // Or you can just declare the `ref` inline in your schema
  *     const postSchema2 = new Schema({
@@ -1011,6 +1044,14 @@ SchemaType.prototype._applySetters = function(value, scope, init, priorVal) {
   return v;
 };
 
+/*!
+ * ignore
+ */
+
+SchemaType.prototype._castNullish = function _castNullish(v) {
+  return v;
+};
+
 /**
  * Applies setters
  *
@@ -1024,7 +1065,7 @@ SchemaType.prototype.applySetters = function(value, scope, init, priorVal, optio
   let v = this._applySetters(value, scope, init, priorVal, options);
 
   if (v == null) {
-    return v;
+    return this._castNullish(v);
   }
 
   // do not cast until all setters are applied #665
@@ -1420,12 +1461,19 @@ SchemaType.prototype.$conditionalHandlers = {
 SchemaType.prototype.castForQueryWrapper = function(params) {
   this.$$context = params.context;
   if ('$conditional' in params) {
-    return this.castForQuery(params.$conditional, params.val);
+    const ret = this.castForQuery(params.$conditional, params.val);
+    this.$$context = null;
+    return ret;
   }
   if (params.$skipQueryCastForUpdate || params.$applySetters) {
-    return this._castForQuery(params.val);
+    const ret = this._castForQuery(params.val);
+    this.$$context = null;
+    return ret;
   }
-  return this.castForQuery(params.val);
+
+  const ret = this.castForQuery(params.val);
+  this.$$context = null;
+  return ret;
 };
 
 /**
